import unittest

from junkie import Junkie, JunkieError


class DictTest(unittest.TestCase):
    def test_getitem(self):
        class App:
            pass

        context = {
            "app": App
        }

        with Junkie(context).inject("app", "_junkie") as (app, junkie):
            self.assertEqual(app, junkie["app"])

    def test_raise_error_on_missing_item(self):
        with self.assertRaisesRegex(KeyError, "'missing'"):
            _ = Junkie()["missing"]

    def test_setitem(self):
        junkie = Junkie()
        junkie["string"] = "my string"
        with junkie.inject("string") as string:
            self.assertEqual("my string", string)

    def test_setitem_for_current_scope_only(self):
        junkie = Junkie()
        with junkie.inject():
            junkie["number"] = 7

            with junkie.inject("number") as number:
                self.assertEqual(7, number)

        self.assertNotIn("number", junkie)

    def test_raise_error_on_setitem_when_overriding(self):
        class App:
            def __init__(self, database):
                self.database = database

        context = {
            "database": "sqlite://"
        }

        with Junkie(context).inject(App, "_junkie") as (app, junkie):
            with self.assertRaisesRegex(JunkieError, '^Instance for "database" already exists in context$'):
                junkie["database"] = "postgres://"

    def test_contains(self):
        class App:
            pass

        context = {
            "app": App
        }

        with Junkie(context).inject("app", "_junkie") as (app, junkie):
            self.assertIn("app", junkie)

    def test_iter(self):
        class App:
            pass

        context = {
            "app": App
        }

        with Junkie(context).inject("app", "_junkie") as (app, junkie):
            self.assertEqual({"app": junkie["app"]}, dict(junkie))



import logging
import random
import sys
import textwrap
import unittest
from contextlib import contextmanager
from unittest import skipIf

from junkie import Junkie, JunkieError, inject_list


class JunkieTest(unittest.TestCase):
    def test_resolve_instance_by_name(self):
        context = {"text": "abc"}

        with Junkie(context).inject("text") as instance:
            self.assertEqual("abc", instance)

    def test_resolve_instance_with_factory_by_name(self):
        context = {"text": lambda: "abc"}

        with Junkie(context).inject("text") as instance:
            self.assertEqual("abc", instance)

    def test_raise_exception_if_instance_name_is_unknown(self):
        with self.assertRaises(Exception) as exception_context:
            with Junkie().inject("instance_name"):
                pass

        self.assertEqual('Unable to find "instance_name"', str(exception_context.exception))

    def test_resolve_instance_with_factory_by_type(self):
        class AppClass:
            def __init__(self, text: str):
                self.text = text

        context = {"text": "abc"}

        with Junkie(context).inject(AppClass) as instance:
            self.assertEqual("abc", instance.text)

    def test_resolve_instance_with_factory_using_two_instances(self):
        context = {
            "prefix": "abc",
            "suffix": "def",
            "text": lambda prefix, suffix: prefix + " " + suffix
        }

        with Junkie(context).inject("text") as text:
            self.assertEqual("abc def", text)

    def test_resolve_instance_parameters(self):
        context = {
            "prefix": "abc",
            "suffix": "def",
            "text": lambda prefix, suffix: prefix + " " + suffix
        }

        with Junkie(context).inject("prefix", "suffix", "text") as (my_prefix, my_suffix, text):
            self.assertEqual(("abc", "def", "abc def"), (my_prefix, my_suffix, text))

    def test_resolve_None_as_parameter(self):
        class Class:
            def __init__(self, empty):
                self.empty = empty

        context = {"empty": None, "class": Class}

        with Junkie(context).inject("empty", "class") as (empty_value, class_value):
            self.assertIsNone(empty_value)
            self.assertIsNone(class_value.empty)

    def test_resolve_tuple_with_correct_order(self):
        def func(letter):
            def factory_func(logger):
                logger.append(letter)
                return letter

            return factory_func

        test_logger = []
        context = {
            "a": func("a"), "b": func("b"), "c": func("c"), "d": func("d"), "e": func("e"),
            "logger": test_logger
        }

        names = ["a", "b", "c", "d", "e"]
        random.shuffle(names)

        with Junkie(context).inject(*names):
            self.assertEqual(names, test_logger)

    def test_default_argument_usage(self):
        class MyClassWithDefaultArguments:
            def __init__(self, argument: str, default_argument: int = 10, default_argument2: str = None):
                self.argument = argument
                self.default_argument = default_argument
                self.default_argument2 = default_argument2 or "Hello"

        context = {"argument": "value"}

        with Junkie(context).inject(MyClassWithDefaultArguments) as instance:
            self.assertEqual("value", instance.argument)
            self.assertEqual(10, instance.default_argument)
            self.assertEqual("Hello", instance.default_argument2)

    def test_partial_default_arguments_usage(self):
        class MyClassWithDefaultArguments:
            def __init__(self, argument: str, default_argument: int = 10, default_argument2: str = None):
                self.argument = argument
                self.default_argument = default_argument
                self.default_argument2 = default_argument2 or "Hello"

        context = {"argument": "value", "default_argument2": "set from context"}

        with Junkie(context).inject(MyClassWithDefaultArguments) as instance:
            self.assertEqual("value", instance.argument)
            self.assertEqual(10, instance.default_argument)
            self.assertEqual("set from context", instance.default_argument2)

    def test_empty_args_usage(self):
        class MyClassWithKwargs:
            def __init__(self, *args):
                self.args = args

        with Junkie().inject(MyClassWithKwargs) as instance:
            self.assertEqual((), instance.args)

    def test_args_usage_with_tuple(self):
        class MyClassWithKwargs:
            def __init__(self, *my_tuple):
                self.my_tuple = my_tuple

        context = {"my_tuple": (1, 2, 3)}

        with Junkie(context).inject(MyClassWithKwargs) as instance:
            self.assertEqual((1, 2, 3), instance.my_tuple)

    def test_args_usage_with_list_as_tuple_input(self):
        class MyClassWithKwargs:
            def __init__(self, *my_tuple):
                self.my_tuple = my_tuple

        context = {"my_tuple": [1, 2, 3]}

        with Junkie(context).inject(MyClassWithKwargs) as instance:
            self.assertEqual((1, 2, 3), instance.my_tuple)

    def test_args_usage_with_factory_function(self):
        class MyClassWithKwargs:
            def __init__(self, *my_tuple):
                self.my_tuple = my_tuple

        context = {"my_tuple": lambda: (1, 2, 3)}

        with Junkie(context).inject(MyClassWithKwargs) as instance:
            self.assertEqual((1, 2, 3), instance.my_tuple)

    def test_empty_kwargs_usage(self):
        class MyClassWithKwargs:
            def __init__(self, **kwargs):
                self.kwargs = kwargs

        with Junkie().inject(MyClassWithKwargs) as instance:
            self.assertEqual({}, instance.kwargs)

    def test_kwargs_usage_with_dictionary(self):
        class MyClassWithKwargs:
            def __init__(self, **my_vars):
                self.my_vars = my_vars

        context = {"my_vars": {"a": "a"}}

        with Junkie(context).inject(MyClassWithKwargs) as instance:
            self.assertEqual({"a": "a"}, instance.my_vars)

    def test_kwargs_usage_with_factory_function(self):
        class MyClassWithKwargs:
            def __init__(self, **my_vars):
                self.my_vars = my_vars

        context = {"my_vars": lambda: {"a": "a"}}

        with Junkie(context).inject(MyClassWithKwargs) as instance:
            self.assertEqual({"a": "a"}, instance.my_vars)

    def test_context_manager_enter_and_exit(self):
        class Class:
            def __init__(self, message_service, database_context):
                self.message_service = message_service
                self.database_context = database_context

        class MessageService:
            def __init__(self, logger):
                self.logger = logger

            def __enter__(self):
                self.logger.append("connect")
                return self

            def __exit__(self, exc_type, exc_val, exc_tb):
                self.logger.append("disconnect")

        @contextmanager
        def create_database(logger):
            logger.append("open")
            yield "DB"
            logger.append("close")

        test_logger = []
        context = {
            "logger": test_logger,
            "message_service": MessageService,
            "database_context": create_database,
        }

        with Junkie(context).inject(Class) as instance:
            self.assertIsInstance(instance.message_service, MessageService)
            self.assertEqual("DB", instance.database_context)

            self.assertEqual(["connect", "open"], test_logger)

        self.assertEqual(["connect", "open", "close", "disconnect"], test_logger)

    @skipIf(sys.version_info < (3, 7), "@dataclass needs at least Python 3.7")
    def test_dataclass_decorator(self):
        exec(textwrap.dedent("""
            from dataclasses import dataclass

            @dataclass
            class Class:
                prefix: str
                suffix: str
                pass

                def __post_init__(self):
                    self.text = self.prefix + " " + self.suffix

            with Junkie({"prefix": "abc", "suffix": "def"}).inject(Class) as instance:
                self.assertEqual("abc def", instance.text)
        """))

    def test_unknown_type_exception(self):
        with self.assertRaises(RuntimeError) as error:
            # noinspection PyTypeChecker
            with Junkie().inject(1):
                pass

        self.assertEqual('Unknown type "1" (str, type or Callable expected)', str(error.exception))

    def test_logging(self):
        self.maxDiff = None

        class Class:
            def __init__(self, database_context, message_service, *args):
                self.database_context = database_context
                self.message_service = message_service
                self.args = args

        @contextmanager
        def create_database(connection_string, **kwargs):
            yield "DB: " + connection_string + str(kwargs)

        @contextmanager
        def create_message_service():
            yield "message-service"

        context = {
            "connection_string": lambda: "URL",
            "database_context": create_database,
            "message_service": create_message_service,
            "class": Class,
            "args": ["app_name"],
            "kwargs": {"engine": "sqlite"},
        }

        with self.assertLogs(level="DEBUG") as logging_context:
            with Junkie(context).inject("class", Class):
                logging.getLogger(__name__).info("execute context block")

        self.assertEqual((
            "DEBUG:inject(class, <class 'test.test_junkie.JunkieTest.test_logging.<locals>.Class'>)\n"
            "DEBUG:connection_string = <lambda>()\n"
            "DEBUG:database_context = create_database(connection_string, engine='sqlite')\n"
            "DEBUG:database_context.__enter__()\n"
            "DEBUG:message_service = create_message_service()\n"
            "DEBUG:message_service.__enter__()\n"
            "DEBUG:class = Class(database_context, message_service, app_name)\n"
            "DEBUG:_ = Class(database_context, message_service, app_name)\n"
            "INFO:test.test_junkie:execute context block\n"
            "DEBUG:message_service.__exit__()\n"
            "DEBUG:database_context.__exit__()"
        ), "\n".join(logging_context.output).replace(":junkie:", ":"))

    def test_auto_inject(self):
        class A:
            pass

        class B:
            def __init__(self, a: A):
                self.a = a

        class C:
            def __init__(self, b: B):
                self.b = b

        with Junkie().inject(C) as c_instance:
            self.assertIsInstance(c_instance, C)
            self.assertIsInstance(c_instance.b, B)
            self.assertIsInstance(c_instance.b.a, A)

    def test_no_auto_inject_for_default_arguments(self):
        class A:
            pass

        class B:
            def __init__(self, a: A = None):
                self.a = a

        with Junkie().inject(B) as b_instance:
            self.assertIsInstance(b_instance, B)
            self.assertIsNone(b_instance.a)

    def test_auto_inject_prioritize_named_from_context(self):
        class A:
            pass

        class B:
            def __init__(self, a: A):
                self.a = a

        context = {"a": "from context"}

        with Junkie(context).inject(B) as b:
            self.assertIsInstance(b, B)
            self.assertEqual("from context", b.a)

    def test_no_create_for_builtins(self):
        with self.assertRaises(JunkieError):
            with Junkie().inject(dict):
                pass

    def test_no_auto_inject_for_builtins(self):
        class B:
            def __init__(self, a: str):
                self.a = a

        with self.assertRaises(JunkieError) as error:
            with Junkie().inject(B):
                pass

        self.assertRegex(str(error.exception), 'Mapping for "a" of builtin type "str" is missing')

    def test_object_is_persisted(self):
        class A:
            pass

        context = {
            "a": A,
        }

        _junkie = Junkie(context)
        with _junkie.inject("a") as a1:
            with _junkie.inject("a") as a2:
                with _junkie.inject("a") as a3:
                    self.assertIs(a1, a2)
                    self.assertIs(a1, a3)

    def test_resolve_instance_per_context_key(self):
        class A:
            pass

        context = {
            "a": A,
            "b": A,
        }

        _junkie = Junkie(context)
        with _junkie.inject("a", "b") as (a1, b1):
            with _junkie.inject("a", "b") as (a2, b2):
                self.assertIsNot(a1, b1)
                self.assertIsNot(a2, b2)
                self.assertIs(a1, a2)
                self.assertIs(b1, b2)

    def test_type_as_key_in_mapping_is_ignored(self):
        class A:
            pass

        context = {
            A: "a",
        }

        # noinspection PyTypeChecker
        with Junkie(context).inject(A) as a:
            self.assertIsInstance(a, A)

    def test_auto_inject_same_instance_by_name(self):
        class A:
            pass

        class A1:
            pass

        class B:
            def __init__(self, a: A):
                self.a = a

        class C:
            def __init__(self, a: A1):
                self.a = a

        with Junkie().inject(B, C) as (b, c):
            self.assertIs(b.a, c.a)
            self.assertIsInstance(b.a, A)

        with Junkie().inject(C, B) as (c, b):
            self.assertIs(b.a, c.a)
            self.assertIsInstance(b.a, A1)

    def test_inject_junkie_reference(self):
        my_junkie = Junkie()

        with my_junkie.inject("_junkie") as injected_junkie:
            self.assertIs(injected_junkie, my_junkie)

    def test_inject_list(self):
        class A:
            pass

        class B:
            def __init__(self, a: A, some_value: str):
                self.a = a
                self.some_value = some_value

        context = {"some_value": "value", "my_list_1": inject_list(A, B), "my_list_2": inject_list(B)}

        with Junkie(context).inject("my_list_1", "my_list_2") as (my_list_1, my_list_2):
            self.assertIsInstance(my_list_1, list)
            self.assertIsInstance(my_list_2, list)
            self.assertIsInstance(my_list_1[0], A)
            self.assertIsInstance(my_list_1[1].a, A)
            self.assertEqual("value", my_list_1[1].some_value)
            self.assertEqual("value", my_list_2[0].some_value)



import sys
import textwrap
import unittest
from unittest import skipIf

from junkie import Junkie


# https://peps.python.org/pep-0570/#specification
class ParameterTest(unittest.TestCase):
    def test_parameters(self):
        class App:
            def __init__(self, positional_or_keyword_parameter, *args, **kwargs):
                self.positional_or_keyword_parameter = positional_or_keyword_parameter
                self.args = args
                self.kwargs = kwargs

        context = {
            "positional_or_keyword_parameter": "positional_or_keyword_parameter",
            "args": ["args"],
            "kwargs": {"kwargs": "kwargs"},
        }

        with Junkie(context).inject(App) as app:
            self.assertEqual("positional_or_keyword_parameter", app.positional_or_keyword_parameter)
            self.assertEqual(("args",), app.args)
            self.assertEqual({"kwargs": "kwargs"}, app.kwargs)

    def test_optional_parameters(self):
        class App:
            def __init__(self, positional_or_keyword_parameter="default", *args, **kwargs):
                self.positional_or_keyword_parameter = positional_or_keyword_parameter
                self.args = args
                self.kwargs = kwargs

        with Junkie().inject(App) as app:
            self.assertEqual("default", app.positional_or_keyword_parameter)
            self.assertEqual((), app.args)
            self.assertEqual({}, app.kwargs)

    @skipIf(sys.version_info < (3, 8), "Positional-only parameters need at least Python 3.8")
    def test_positional_and_keyword_parameters(self):
        exec(textwrap.dedent("""
            class App:
                def __init__(
                        self,
                        positional_only_parameter,
                        /,
                        positional_or_keyword_parameter,
                        *,
                        keyword_only_parameter
                ):
                    self.positional_only_parameter = positional_only_parameter
                    self.positional_or_keyword_parameter = positional_or_keyword_parameter
                    self.keyword_only_parameter = keyword_only_parameter
    
            context = {
                "positional_only_parameter": "positional_only_parameter",
                "positional_or_keyword_parameter": "positional_or_keyword_parameter",
                "keyword_only_parameter": "keyword_only_parameter",
            }
    
            with Junkie(context).inject(App) as app:
                self.assertEqual("positional_only_parameter", app.positional_only_parameter)
                self.assertEqual("positional_or_keyword_parameter", app.positional_or_keyword_parameter)
                self.assertEqual("keyword_only_parameter", app.keyword_only_parameter)
        """))

    @skipIf(sys.version_info < (3, 8), "Positional-only parameters need at least Python 3.8")
    def test_all_default_values(self):
        exec(textwrap.dedent("""
            class App:
                def __init__(
                        self,
                        positional_only_parameter="default_positional_only",
                        /,
                        positional_or_keyword_parameter="default_positional_or_keyword",
                        *,
                        keyword_only_parameter="default_keyword_only",
                ):
                    self.positional_only_parameter = positional_only_parameter
                    self.positional_or_keyword_parameter = positional_or_keyword_parameter
                    self.keyword_only_parameter = keyword_only_parameter
    
            with Junkie().inject(App) as app:
                self.assertEqual("default_positional_only", app.positional_only_parameter)
                self.assertEqual("default_positional_or_keyword", app.positional_or_keyword_parameter)
                self.assertEqual("default_keyword_only", app.keyword_only_parameter)
    
            context = {
                "positional_only_parameter": "positional_only_parameter",
                "positional_or_keyword_parameter": "positional_or_keyword_parameter",
                "keyword_only_parameter": "keyword_only_parameter",
            }
    
            with Junkie(context).inject(App) as app:
                self.assertEqual("positional_only_parameter", app.positional_only_parameter)
                self.assertEqual("positional_or_keyword_parameter", app.positional_or_keyword_parameter)
                self.assertEqual("keyword_only_parameter", app.keyword_only_parameter)
        """))



import builtins
import inspect
import logging
from collections import OrderedDict
from contextlib import contextmanager, ExitStack
from typing import Union, Tuple, Mapping, Any, Callable, Optional, Iterator

import junkie

LOGGER = logging.getLogger(junkie.__name__)
BUILTINS = {item for item in vars(builtins).values() if isinstance(item, type)}


def get_factory_name(factory: Callable) -> str:
    return getattr(factory, "__name__", str(factory))


class JunkieError(RuntimeError):
    pass


class Junkie(Mapping[str, Any]):
    def __init__(self, instances_and_factories: Mapping[str, Any] = None):
        self._context = instances_and_factories or {}
        self._exit_stack: Optional[ExitStack] = None

        self._instances_by_name: dict = {}
        self._instances_by_name_stack: Junkie._Stack = Junkie._Stack()
        self._instances_by_name_stack.push(self._instances_by_name)

        self._instantiation_stack: Junkie._InstantiationStack = Junkie._InstantiationStack()
        self._cycle_detection_instance_set = set()

        self._context["_junkie"] = self

    def __getitem__(self, item):
        return self._instances_by_name[item]

    def __setitem__(self, key, value):
        if key in self._instances_by_name:
            raise JunkieError(f'Instance for "{key}" already exists')
        if key in self._context:
            raise JunkieError(f'Instance for "{key}" already exists in context')

        self._instances_by_name[key] = value

    def __len__(self) -> int:
        return self._instances_by_name.__len__()

    def __iter__(self) -> Iterator:
        return self._instances_by_name.__iter__()

    def __contains__(self, item) -> bool:
        return self._instances_by_name.__contains__(item)

    @contextmanager
    def inject(self, *names_and_factories: Union[str, Callable]) -> Union[Any, Tuple[Any]]:
        LOGGER.debug("inject(%s)", Junkie._LogParams(*names_and_factories))

        with ExitStack() as self._exit_stack:
            self._instances_by_name = self._instances_by_name_stack.peek().copy()
            self._instances_by_name_stack.push(self._instances_by_name)

            if len(names_and_factories) == 1:
                yield self._build_instance(names_and_factories[0])
            else:
                yield self._build_tuple(*names_and_factories)

            self._instances_by_name_stack.pop()
            self._instances_by_name = self._instances_by_name_stack.peek()

    def _build_tuple(self, *names_and_factories: Union[str, Callable]) -> Tuple[Any, ...]:
        instances = []

        for name_or_factory in names_and_factories:
            instance = self._build_instance(name_or_factory)
            instances.append(instance)

        return tuple(instances)

    def _build_instance(self, name_or_factory: Union[str, Callable]) -> Any:
        if isinstance(name_or_factory, str):
            return self._build_by_instance_name(name_or_factory)

        elif callable(name_or_factory):
            return self._build_by_factory_function(name_or_factory, None)

        raise JunkieError(
            f"{self._instantiation_stack}" + f'Unknown type "{name_or_factory}" (str, type or Callable expected)'
        )

    def _build_by_instance_name(self, instance_name: str) -> Any:
        if instance_name in self._instances_by_name:
            return self._instances_by_name[instance_name]

        if instance_name in self._context:
            value = self._context[instance_name]

            if callable(value):
                return self._build_by_factory_function(value, instance_name)
            else:
                return value

        raise JunkieError(f"{self._instantiation_stack}" + f'Unable to find "{instance_name}"')

    def _build_by_factory_function(self, factory_function: Callable, instance_name: Union[str, None]) -> Any:
        if factory_function in BUILTINS:
            raise JunkieError(
                f"{self._instantiation_stack}"
                + f'Mapping for "{instance_name}" of builtin type "{get_factory_name(factory_function)}" is missing'
            )

        if factory_function in self._cycle_detection_instance_set:
            raise JunkieError(
                f"{self._instantiation_stack}"
                + f'Dependency cycle detected with "{get_factory_name(factory_function)}()"'
            )

        self._cycle_detection_instance_set.add(factory_function)
        self._instantiation_stack.push(factory_function)

        positional_params, args, keyword_params, kwargs = self._build_parameters(factory_function)

        if LOGGER.isEnabledFor(logging.DEBUG):
            log_params = Junkie._LogParams(*positional_params.keys(), *args, **keyword_params, **kwargs)
            LOGGER.debug("%s = %s(%s)", instance_name or "_", get_factory_name(factory_function), log_params)

        instance = factory_function(*positional_params.values(), *args, **keyword_params, **kwargs)

        if hasattr(instance, "__enter__"):
            if LOGGER.isEnabledFor(logging.DEBUG):
                LOGGER.debug("%s.__enter__()", instance_name or "_")
                self._exit_stack.push(lambda *exception_details: LOGGER.debug("%s.__exit__()", instance_name or "_"))

            instance = self._exit_stack.enter_context(instance)

        if instance_name is not None:
            self._instances_by_name[instance_name] = instance

        self._instantiation_stack.pop()
        self._cycle_detection_instance_set.remove(factory_function)

        return instance

    def _build_parameters(self, factory_function: Callable) -> (OrderedDict, tuple, dict):
        positional_params = OrderedDict()
        args = ()
        keyword_params = OrderedDict()
        kwargs = {}
        positional_params_finished = False

        try:
            signature = inspect.signature(factory_function)
        except Exception as e:
            raise JunkieError(
                f"{self._instantiation_stack}"
                + f'Unable to inspect signature for "{get_factory_name(factory_function)}()"'
            ) from e

        for instance_name, annotation in signature.parameters.items():
            if instance_name in self._instances_by_name or instance_name in self._context:
                value = self._build_by_instance_name(instance_name)

            # *args
            elif annotation.kind is inspect.Parameter.VAR_POSITIONAL:
                continue

            # **kwargs
            elif annotation.kind is inspect.Parameter.VAR_KEYWORD:
                continue

            # arg="value"
            elif annotation.default is not inspect.Parameter.empty:
                positional_params_finished = True
                continue

            elif isinstance(annotation.annotation, Callable) and annotation.annotation != inspect.Parameter.empty:
                value = self._build_by_factory_function(annotation.annotation, instance_name)

            else:
                raise JunkieError(
                    f"{self._instantiation_stack}"
                    + f'Unable to find "{instance_name}" for "{get_factory_name(factory_function)}()"'
                )

            if annotation.kind is inspect.Parameter.POSITIONAL_ONLY:
                positional_params[instance_name] = value

            elif annotation.kind is inspect.Parameter.POSITIONAL_OR_KEYWORD:
                if positional_params_finished:
                    keyword_params[instance_name] = value
                else:
                    positional_params[instance_name] = value

            elif annotation.kind is inspect.Parameter.VAR_POSITIONAL:
                args = value

            elif annotation.kind is inspect.Parameter.KEYWORD_ONLY:
                keyword_params[instance_name] = value

            elif annotation.kind is inspect.Parameter.VAR_KEYWORD:
                kwargs = value

            else:
                raise NotImplementedError(f'Unknown parameter type "{annotation.kind}"')

        return positional_params, args, keyword_params, kwargs

    class _LogParams:
        def __init__(self, *args, **kwargs):
            self.args = args
            self.kwargs = kwargs

        def __str__(self):
            arg_params = list(map(str, self.args))
            kwarg_params = list(map(str, [f"{key}={repr(value)}" for key, value in self.kwargs.items()]))
            return ", ".join(arg_params + kwarg_params)

    class _Stack:
        def __init__(self):
            self._stack = []

        def push(self, item):
            self._stack.append(item)

        def pop(self):
            return self._stack.pop()

        def peek(self):
            return self._stack[-1]

        def __len__(self):
            return self._stack.__len__()

    class _InstantiationStack(_Stack):
        def __str__(self):
            if len(self._stack) == 0:
                return ""

            return "".join([
                f'\n{idx * " "}-> {get_factory_name(factory)}() at {self._get_source_info(factory)}'
                for idx, factory in enumerate(self._stack)
            ]) + "\n"

        @staticmethod
        def _get_source_info(factory: Callable) -> str:
            while hasattr(factory, "__wrapped__"):
                factory = factory.__wrapped__
            try:
                return f'"{inspect.getsourcefile(factory)}:{inspect.getsourcelines(factory)[1]}"'
            except:
                return "unknown source"


def inject_list(*factories_or_names):
    """Can be used within the context to let junkie create a list of instances from a list of factories or names"""

    @contextmanager
    def wrapper(_junkie: Junkie):
        with _junkie.inject(*factories_or_names) as instances:
            if isinstance(instances, tuple):
                yield list(instances)
            else:
                yield [instances]

    return wrapper



import sqlite3
import unittest
from contextlib import contextmanager
from functools import lru_cache

from junkie import Junkie, JunkieError, inject_list


class ErrorTest(unittest.TestCase):
    def test_list_class_in_error_message(self):
        class Database:
            def __init__(self, _):
                pass

        class App:
            def __init__(self, database: Database):
                self.database = database

        message = (
            "\n"
            r'-> App\(\) at ".*/test/test_error.py:\d+"' '\n'
            r' -> Database\(\) at ".*/test/test_error.py:\d+"' '\n'
            r'Unable to find "_" for "Database\(\)"'
        )

        with self.assertRaisesRegex(JunkieError, message):
            with Junkie().inject(App):
                pass

    def test_list_function_in_error_message(self):
        def connect_database(_):
            pass

        class App:
            def __init__(self, database: connect_database):
                self.database = database

        message = (
            "\n"
            r'-> App\(\) at ".*/test/test_error.py:\d+"' '\n'
            r' -> connect_database\(\) at ".*/test/test_error.py:\d+"' '\n'
            r'Unable to find "_" for "connect_database\(\)"'
        )

        with self.assertRaisesRegex(JunkieError, message):
            with Junkie().inject(App):
                pass

    def test_list_wrapped_function_in_error_message(self):
        @lru_cache(3)
        @contextmanager
        def connect_database(_):
            yield

        class App:
            def __init__(self, database):
                self.database = database

        context = {
            "database": connect_database
        }

        message = (
            "\n"
            r'-> App\(\) at ".*/test/test_error.py:\d+"' '\n'
            r' -> connect_database\(\) at ".*/test/test_error.py:\d+"' '\n'
            r'Unable to find "_" for "connect_database\(\)"'
        )

        with self.assertRaisesRegex(JunkieError, message):
            with Junkie(context).inject(App):
                pass

    def test_list_lambda_in_error_message(self):
        class App:
            def __init__(self, database):
                self.database = database

        context = {
            "database": lambda unknown: unknown
        }

        message = (
            "\n"
            r'-> App\(\) at ".*/test/test_error.py:\d+"' '\n'
            r' -> <lambda>\(\) at ".*/test/test_error.py:\d+"' '\n'
            r'Unable to find "unknown" for "<lambda>\(\)"'
        )

        with self.assertRaisesRegex(JunkieError, message):
            with Junkie(context).inject(App):
                pass

    def test_list_unknown_in_error_message(self):
        class App:
            def __init__(self, database):
                self.database = database

        context = {
            "database": sqlite3.connect
        }

        message = (
            "\n"
            r'-> App\(\) at ".*/test/test_error.py:\d+"' '\n'
            r' -> connect\(\) at unknown source' '\n'
            r'Unable to inspect signature for "connect\(\)"'
        )

        with self.assertRaisesRegex(JunkieError, message):
            with Junkie(context).inject(App):
                pass

    def test_one_stack_trace_in_error_message(self):
        class App:
            def __init__(self, database):
                self.database = database

        context = {
            "database": inject_list("file"),
            "file": lambda fs: fs.file,
        }

        message = (
            "\n"
            r'-> App\(\) at ".*/test/test_error.py:\d+"' '\n'
            r' -> wrapper\(\) at ".*/junkie/_junkie.py:\d+"' '\n'
            r'  -> <lambda>\(\) at ".*/test/test_error.py:\d+"' '\n'
            r'Unable to find "fs" for "<lambda>\(\)"'
        )

        with self.assertRaisesRegex(JunkieError, message):
            with Junkie(context).inject(App):
                pass

    def test_detect_dependency_cycle(self):
        class App:
            def __init__(self, app):
                app()

        context = {"app": App}

        with self.assertRaisesRegex(JunkieError, r'.*Dependency cycle detected with "App\(\)"'):
            with Junkie(context).inject(App):
                pass



import unittest


class ReadmeTest(unittest.TestCase):
    def test_example(self):
        from junkie import Junkie

        class App:
            def __init__(self, addressee):
                self.addressee = addressee

            def greets(self):
                return f"Hello {self.addressee}!"

        context = {"addressee": "World"}

        with Junkie(context).inject(App) as app:
            assert app.greets() == "Hello World!"

    def test_type_hints(self):
        from junkie import Junkie

        class Database:
            pass

        class QueryHelper:
            def __init__(self, database: Database):
                self.database = database

        class App:
            def __init__(self, database: Database, query_helper: QueryHelper):
                self.database = database
                self.query_helper = query_helper

        with Junkie().inject(App) as app:  # type: App
            assert isinstance(app.database, Database)
            assert app.query_helper.database == app.database

    def test_integration_test(self):
        from junkie import Junkie

        APPLICATION_CONTEXT = {
            "database_url": "postgresql://scott:tiger@localhost:5432/production",
        }

        class App:
            def __init__(self, database_url):
                self.database_url = database_url

        test_context = {**APPLICATION_CONTEXT, "database_url": "sqlite://"}

        with Junkie(test_context).inject(App) as app:
            self.assertEqual(app.database_url, "sqlite://")

    def test_reuse(self):
        from junkie import Junkie

        class App:
            pass

        context = {
            "app": App,
        }

        with Junkie(context).inject("app", App, "app") as (app1, app2, app3):
            assert app1 == app3
            assert app1 != app2 != app3

    def test_lambdas(self):
        from junkie import Junkie

        class App:
            def __init__(self, greeting: str):
                self.greeting = greeting

        context = {
            # app1
            "app1": lambda: App("Hello Joe!"),
            # app2
            "greeting2": "Hello John!",
            "app2": lambda greeting2: App(greeting2),
            # app3
            "greeting3": lambda: "Hello Doe!",
            "app3": lambda greeting3: App(greeting3),
        }

        with Junkie(context).inject("app1", "app2", "app3") as (app1, app2, app3):
            assert app1.greeting == "Hello Joe!"
            assert app2.greeting == "Hello John!"
            assert app3.greeting == "Hello Doe!"

    def test_junkie_argument(self):
        from contextlib import contextmanager

        from junkie import Junkie

        class SqlDatabase:
            pass

        class FileDatabase:
            pass

        class App:
            def __init__(self, database):
                self.database = database

        @contextmanager
        def provide_database(_junkie, url: str):
            if url.startswith("file:"):
                with _junkie.inject(FileDatabase) as database:
                    yield database
            else:
                with _junkie.inject(SqlDatabase) as database:
                    yield database

        context = {
            "url": "file://local.db",
            "database": provide_database,
        }

        with Junkie(context).inject(App) as app:
            assert isinstance(app.database, FileDatabase)

    def test_inject_list(self):
        from junkie import Junkie, inject_list

        class CustomerDataSource:
            def __init__(self, connection_string: str):
                pass

        class ProductDataSource:
            pass

        class SupplierDataSource:
            pass

        class App:
            def __init__(self, data_sources):
                self.data_sources = data_sources

        context = {
            "customer_ds": lambda: CustomerDataSource("sqlite://"),
            "data_sources": inject_list("customer_ds", ProductDataSource, SupplierDataSource),
        }

        with Junkie(context).inject(App) as app:
            assert isinstance(app.data_sources[0], CustomerDataSource)
            assert isinstance(app.data_sources[1], ProductDataSource)
            assert isinstance(app.data_sources[2], SupplierDataSource)

    def test_pre_defined_callable(self):
        from junkie import Junkie

        class Database:
            def __call__(self, *args, **kwargs):
                return "called"

        class App:
            def __init__(self, database):
                self.database = database

        context = {
            "database": lambda: Database(),
        }

        with Junkie(context).inject(App) as app:
            assert app.database() == "called"

    def test_lambda_for_builtins(self):
        import sqlite3

        from junkie import Junkie, JunkieError

        context = {
            "database": ":memory:",
            "connection": sqlite3.connect,
            "working_connection": lambda database: sqlite3.connect(database)
        }

        try:
            # ValueError: no signature found for builtin <built-in function connect>
            with Junkie(context).inject("connection") as connection:
                pass
            assert False
        except JunkieError:
            pass

        with Junkie(context).inject("working_connection") as working_connection:
            pass



import logging
import unittest
from unittest.mock import Mock

from junkie import Junkie, JunkieError


class BugfixTest(unittest.TestCase):
    def test_error_on_missing_mapping(self):
        class App:
            def __init__(self, _):
                pass

        with self.assertRaises(JunkieError):
            with Junkie().inject(App):
                pass

    def test_no_error_for_callable_without_name(self):
        context = {
            "mock": Mock()
        }
        with self.assertLogs(level=logging.DEBUG) as log:
            with Junkie(context).inject("mock"):
                pass

        self.assertRegex(str(log.output), r"DEBUG:junkie:mock = <Mock id='\d+'>\(\)")



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='junkie',
    version='3.0.1',
    packages=setuptools.find_packages(exclude="test"),
    author='Stefan Richter',
    description='A dependency injection library for beginners and professionals',
    url="https://github.com/sealor/junkie",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)
